# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import tempfile
import collections
import os
import subprocess
import time

from pymor.analyticalproblems.domaindescriptions import RectDomain, PolygonalDomain
from pymor.discretizers.builtin.grids.gmsh import load_gmsh
from pymor.core.exceptions import GmshMissing
from pymor.core.logger import getLogger


def discretize_gmsh(domain_description=None, geo_file=None, geo_file_path=None, msh_file_path=None,
                    mesh_algorithm='del2d', clscale=1., options='', refinement_steps=0):
    """Mesh a |DomainDescription| or an already existing Gmsh GEO-file using the Gmsh mesher.

    Parameters
    ----------
    domain_description
        A |DomainDescription| of the |PolygonalDomain| or |RectDomain| to discretize. Has to be
        `None` when `geo_file` is given.
    geo_file
        File handle of the Gmsh Geo-file to discretize. Has to be `None` when
        `domain_description` is given.
    geo_file_path
        Path of the created Gmsh GEO-file. When meshing a |PolygonalDomain| or |RectDomain| and
        `geo_file_path` is `None`, a temporary file will be created. If `geo_file` is specified,
        this is ignored and the path to `geo_file` will be used.
    msh_file_path
        Path of the created Gmsh MSH-file. If `None`, a temporary file will be created.
    mesh_algorithm
        The mesh generation algorithm to use (meshadapt, del2d, front2d).
    clscale
        Mesh element size scaling factor.
    options
        Other options to control the meshing procedure of Gmsh. See
        http://geuz.org/gmsh/doc/texinfo/gmsh.html#Command_002dline-options for all available
        options.
    refinement_steps
        Number of refinement steps to do after the initial meshing.

    Returns
    -------
    grid
        The generated :class:`~pymor.discretizers.builtin.grids.gmsh.GmshGrid`.
    boundary_info
        The generated :class:`~pymor.discretizers.builtin.grids.gmsh.GmshBoundaryInfo`.
    """
    assert domain_description is None or geo_file is None
    logger = getLogger('pymor.discretizers.builtin.domaindiscretizers.gmsh.discretize_gmsh')

    # run Gmsh; initial meshing
    logger.info('Checking for Gmsh ...')

    # when we are running MPI parallel and Gmsh is compiled with MPI support,
    # we have to make sure Gmsh does not notice the MPI environment or it will fail.
    env = {k: v for k, v in os.environ.items()
           if 'MPI' not in k.upper()}
    try:
        version = subprocess.check_output(['gmsh', '--version'], stderr=subprocess.STDOUT, env=env).decode()
    except (subprocess.CalledProcessError, OSError):
        raise GmshMissing('Could not find Gmsh.'
                          ' Please ensure that the gmsh binary (http://geuz.org/gmsh/) is in your PATH.')

    logger.info('Found version ' + version.strip())

    def discretize_PolygonalDomain():
        # combine points and holes, since holes are points, too, and have to be stored as such.
        points = [domain_description.points]
        points.extend(domain_description.holes)

        return points, domain_description.boundary_types

    def discretize_RectDomain():
        points = [[domain_description.domain[0].tolist(),
                   [domain_description.domain[1][0], domain_description.domain[0][1]],
                   domain_description.domain[1].tolist(),
                   [domain_description.domain[0][0], domain_description.domain[1][1]]]]
        boundary_types = {domain_description.bottom: [1]}
        if domain_description.right not in boundary_types:
            boundary_types[domain_description.right] = [2]
        else:
            boundary_types[domain_description.right].append(2)
        if domain_description.top not in boundary_types:
            boundary_types[domain_description.top] = [3]
        else:
            boundary_types[domain_description.top].append(3)
        if domain_description.left not in boundary_types:
            boundary_types[domain_description.left] = [4]
        else:
            boundary_types[domain_description.left].append(4)

        if None in boundary_types:
            del boundary_types[None]

        return points, boundary_types

    # these two are referenced in a finally block, but were left undefined in some paths
    geo_file, msh_file = None, None
    try:
        # When a |PolygonalDomain| or |RectDomain| has to be discretized create a
        # Gmsh GE0-file and write all data.
        if domain_description is not None:
            logger.info('Writing Gmsh geometry file ...')
            # Create a temporary GEO-file if None is specified
            if geo_file_path is None:
                geo_file = tempfile.NamedTemporaryFile(mode='wt', delete=False, suffix='.geo')
                geo_file_path = geo_file.name
            else:
                geo_file = open(geo_file_path, 'w')

            if isinstance(domain_description, PolygonalDomain):
                points, boundary_types = discretize_PolygonalDomain()
            elif isinstance(domain_description, RectDomain):
                points, boundary_types = discretize_RectDomain()
            else:
                raise NotImplementedError(f'I do not know how to discretize {domain_description}')

            # assign ids to all points and write them to the GEO-file.
            for id, p in enumerate([p for ps in points for p in ps]):
                assert len(p) == 2
                geo_file.write('Point('+str(id+1)+') = '+str(p+[0, 0]).replace('[', '{').replace(']', '}')+';\n')

            # store points and their ids
            point_ids = dict(zip([str(p) for ps in points for p in ps],
                                 range(1, len([p for ps in points for p in ps])+1)))
            # shift points 1 entry to the left.
            points_deque = [collections.deque(ps) for ps in points]
            for ps_d in points_deque:
                ps_d.rotate(-1)
            # create lines by connecting the points with shifted points,
            # such that they form a polygonal chains.
            lines = [[point_ids[str(p0)], point_ids[str(p1)]]
                     for ps, ps_d in zip(points, points_deque) for p0, p1 in zip(ps, ps_d)]
            # assign ids to all lines and write them to the GEO-file.
            for l_id, l in enumerate(lines):
                geo_file.write('Line('+str(l_id+1)+')'+' = '+str(l).replace('[', '{').replace(']', '}')+';\n')

            # form line_loops (polygonal chains), create ids and write them to file.
            line_loops = [[point_ids[str(p)] for p in ps] for ps in points]
            line_loop_ids = range(len(lines)+1, len(lines)+len(line_loops)+1)
            for ll_id, ll in zip(line_loop_ids, line_loops):
                geo_file.write('Line Loop('+str(ll_id)+')'+' = '+str(ll).replace('[', '{').replace(']', '}')+';\n')

            # set this here explicitly for string conversion to make sense
            line_loop_ids = list(line_loop_ids)
            # create the surface defined by line loops,
            # starting with the exterior and then the holes.
            geo_file.write('Plane Surface(' + str(line_loop_ids[0]+1) + ')' + ' = '
                           + str(line_loop_ids).replace('[', '{').replace(']', '}') + ';\n')
            geo_file.write('Physical Surface("boundary") = {'+str(line_loop_ids[0]+1)+'};\n')

            # write boundaries.
            for boundary_type, bs in boundary_types.items():
                geo_file.write('Physical Line' + '("' + str(boundary_type) + '")' + ' = '
                               + str([l_id for l_id in bs]).replace('[', '{').replace(']', '}') + ';\n')

            geo_file.close()
        # When a GEO-File is provided just get the corresponding file path.
        else:
            geo_file_path = geo_file.name
        # Create a temporary MSH-file if no path is specified.
        if msh_file_path is None:
            msh_file = tempfile.NamedTemporaryFile(mode='wt', delete=False, suffix='.msh')
            msh_file_path = msh_file.name
            msh_file.close()

        tic = time.perf_counter()

        # run Gmsh; initial meshing
        logger.info('Calling Gmsh ...')
        cmd = ['gmsh', geo_file_path, '-2', '-algo', mesh_algorithm, '-clscale', str(clscale), options, '-o',
               msh_file_path]
        subprocess.check_call(cmd, env=env)

        # run gmsh; perform mesh refinement
        cmd = ['gmsh', msh_file_path, '-refine', '-o', msh_file_path]
        for i in range(refinement_steps):
            logger.info(f'Performing Gmsh refinement step {i+1}')
            subprocess.check_call(cmd, env=env)

        toc = time.perf_counter()
        t_gmsh = toc - tic
        logger.info(f'Gmsh took {t_gmsh} s')

        # Create |GmshGrid| and |GmshBoundaryInfo| form the just created MSH-file.
        grid, bi = load_gmsh(msh_file_path)
    finally:
        # delete tempfiles if they were created beforehand.
        if isinstance(geo_file, tempfile._TemporaryFileWrapper):
            os.remove(geo_file_path)
        if isinstance(msh_file, tempfile._TemporaryFileWrapper):
            os.remove(msh_file_path)

    return grid, bi



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import weakref

import numpy as np

from pymor.discretizers.builtin.grids.boundaryinfos import GenericBoundaryInfo
from pymor.discretizers.builtin.grids.interfaces import Grid


class SubGrid(Grid):
    """A subgrid of a |Grid|.

    Given a |Grid| and a list of codim-0 entities we construct the minimal
    subgrid of the grid, containing all the given entities.

    Parameters
    ----------
    parent_grid
        |Grid| of which a subgrid is to be created.
    parent_entities
        |NumPy array| of global indices of the codim-0 entities which
        are to be contained in the subgrid.

    Attributes
    ----------
    parent_grid
        The |Grid| from which the subgrid was constructed. :class:`Subgrid`
        only stores a :mod:`weakref` to the grid, so accessing this property
        might return `None` if the original grid has been destroyed.
    """

    reference_element = None

    def __init__(self, parent_grid, parent_entities):
        assert parent_grid is not None, \
            'parent_grid is None. Maybe you have called sub_grid.with(parent_entities=e)\n' \
            'on a SubGrid for which the parent grid has been destroyed?'
        assert isinstance(parent_grid, Grid)
        self.dim = parent_grid.dim
        self.reference_element = parent_grid.reference_element

        parent_indices = [np.array(np.unique(parent_entities), dtype=np.int32)]
        assert len(parent_indices[0] == len(parent_entities))

        subentities = [np.arange(len(parent_indices[0]), dtype=np.int32).reshape((-1, 1))]

        for codim in range(1, self.dim + 1):
            SUBE = parent_grid.subentities(0, codim)[parent_indices[0]]
            if np.any(SUBE < 0):
                raise NotImplementedError
            UI, UI_inv = np.unique(SUBE, return_inverse=True)
            subentities.append(np.array(UI_inv.reshape(SUBE.shape), dtype=np.int32))
            parent_indices.append(np.array(UI, dtype=np.int32))

        self.parent_entities = parent_entities
        self.__parent_grid = weakref.ref(parent_grid)
        self.__parent_indices = parent_indices
        self.__subentities = subentities
        embeddings = parent_grid.embeddings(0)
        self.__embeddings = (embeddings[0][parent_indices[0]], embeddings[1][parent_indices[0]])

    @property
    def parent_grid(self):
        return self.__parent_grid()

    def parent_indices(self, codim):
        """`retval[e]` is the index of the `e`-th codim-`codim` entity in the parent grid."""
        assert 0 <= codim <= self.dim, 'Invalid codimension'
        return self.__parent_indices[codim]

    def indices_from_parent_indices(self, ind, codim):
        """Maps indices of codim-`codim` entities of the parent grid to indices of the subgrid.

        Parameters
        ----------
        ind
            |NumPy array| of indices of codim-`codim` entities of the parent grid
        codim
            codim of the entities indicated by `ind`

        Raises
        ------
        ValueError
            Not all provided indices correspond to entities contained in the subgrid.
        """
        assert 0 <= codim <= self.dim, 'Invalid codimension'
        ind = ind.ravel()
        # TODO Find better implementation of the following
        R = np.argmax(ind[:, np.newaxis] - self.__parent_indices[codim][np.newaxis, :] == 0, axis=1)
        if not np.all(self.__parent_indices[codim][R] == ind):
            raise ValueError('Not all parent indices found')
        return np.array(R, dtype=np.int32)

    def size(self, codim):
        assert 0 <= codim <= self.dim, 'Invalid codimension'
        return len(self.__parent_indices[codim])

    def subentities(self, codim, subentity_codim):
        if codim == 0:
            assert codim <= subentity_codim <= self.dim, 'Invalid subentity codimension'
            return self.__subentities[subentity_codim]
        else:
            return super().subentities(codim, subentity_codim)

    def embeddings(self, codim):
        if codim == 0:
            return self.__embeddings
        else:
            return super().embeddings(codim)

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_SubGrid__parent_grid']
        return d


def make_sub_grid_boundary_info(sub_grid, parent_grid, parent_grid_boundary_info, new_boundary_type=None):
    """Derives a |BoundaryInfo| for a :class:`~pymor.discretizers.builtin.grids.subgrid.SubGrid`.

    Parameters
    ----------
    sub_grid
        The :class:`~pymor.discretizers.builtin.grids.subgrid.SubGrid` for which a
        |BoundaryInfo| is created.
    parent_grid
        The parent |Grid|.
    parent_grid_boundary_info
        The |BoundaryInfo| of the parent |Grid| from which to derive the |BoundaryInfo|
    new_boundary_type
        The boundary type which is assigned to the new boundaries of `subgrid`. If
        `None`, no boundary type is assigned.

    Returns
    -------
    |BoundaryInfo| associated with sub_grid.
    """
    boundary_types = parent_grid_boundary_info.boundary_types

    masks = {}
    for codim in range(1, sub_grid.dim + 1):
        parent_indices = sub_grid.parent_indices(codim)[sub_grid.boundaries(codim)]
        new_boundaries = np.where(np.logical_not(parent_grid.boundary_mask(codim)[parent_indices]))
        new_boundaries_sg_indices = sub_grid.boundaries(codim)[new_boundaries]
        for t in boundary_types:
            m = parent_grid_boundary_info.mask(t, codim)[sub_grid.parent_indices(codim)]
            if t == new_boundary_type:
                m[new_boundaries_sg_indices] = True
            if codim == 1:
                masks[t] = [m]
            else:
                masks[t].append(m)
        if new_boundary_type is not None and new_boundary_type not in boundary_types:
            m = np.zeros(sub_grid.size(codim), dtype=np.bool)
            m[new_boundaries_sg_indices] = True
            if codim == 1:
                masks[new_boundary_type] = [m]
            else:
                masks[new_boundary_type].append(m)

    return GenericBoundaryInfo(sub_grid, masks)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np

from pymor.algorithms.rules import RuleTable, match_class, match_generic, match_always
from pymor.core.exceptions import RuleNotMatchingError, NoMatchingRuleError
from pymor.operators.block import BlockOperatorBase, BlockRowOperator, BlockColumnOperator
from pymor.operators.constructions import (LincombOperator, ConcatenationOperator, ConstantOperator, ProjectedOperator,
                                           ZeroOperator, AffineOperator, AdjointOperator, SelectionOperator,
                                           IdentityOperator, VectorArrayOperator)
from pymor.operators.ei import EmpiricalInterpolatedOperator, ProjectedEmpiciralInterpolatedOperator
from pymor.operators.numpy import NumpyMatrixOperator
from pymor.vectorarrays.numpy import NumpyVectorSpace


def project(op, range_basis, source_basis, product=None):
    """Petrov-Galerkin projection of a given |Operator|.

    Given an inner product `( ⋅, ⋅)`, source vectors `b_1, ..., b_N`
    and range vectors `c_1, ..., c_M`, the projection `op_proj` of `op`
    is defined by ::

        [ op_proj(e_j) ]_i = ( c_i, op(b_j) )

    for all i,j, where `e_j` denotes the j-th canonical basis vector of R^N.

    In particular, if the `c_i` are orthonormal w.r.t. the given product,
    then `op_proj` is the coordinate representation w.r.t. the `b_i/c_i` bases
    of the restriction of `op` to `span(b_i)` concatenated with the
    orthogonal projection onto `span(c_i)`.

    From another point of view, if `op` is viewed as a bilinear form
    (see :meth:`apply2`) and `( ⋅, ⋅ )` is the Euclidean inner
    product, then `op_proj` represents the matrix of the bilinear form restricted
    to `span(b_i) / span(c_i)` (w.r.t. the `b_i/c_i` bases).

    How the projection is realized will depend on the given |Operator|.
    While a projected |NumpyMatrixOperator| will
    again be a |NumpyMatrixOperator|, only a generic
    :class:`~pymor.operators.constructions.ProjectedOperator` can be returned
    in general. The exact algorithm is specified in :class:`ProjectRules`.

    Parameters
    ----------
    range_basis
        The vectors `c_1, ..., c_M` as a |VectorArray|. If `None`, no
        projection in the range space is performed.
    source_basis
        The vectors `b_1, ..., b_N` as a |VectorArray| or `None`. If `None`,
        no restriction of the source space is performed.
    product
        An |Operator| representing the inner product.  If `None`, the
        Euclidean inner product is chosen.

    Returns
    -------
    The projected |Operator| `op_proj`.
    """
    assert source_basis is None or source_basis in op.source
    assert range_basis is None or range_basis in op.range
    assert product is None or product.source == product.range == op.range

    rb = product.apply(range_basis) if product is not None and range_basis is not None else range_basis

    try:
        return ProjectRules(rb, source_basis).apply(op)
    except NoMatchingRuleError:
        op.logger.warning('Using inefficient generic projection operator')
        return ProjectedOperator(op, range_basis, source_basis, product)


class ProjectRules(RuleTable):
    """|RuleTable| for the :func:`project` algorithm."""

    def __init__(self, range_basis, source_basis):
        super().__init__(use_caching=True)
        self.__auto_init(locals())

    @match_always
    def action_no_bases(self, op):
        if self.range_basis is None and self.source_basis is None:
            return op
        else:
            raise RuleNotMatchingError

    @match_class(ZeroOperator)
    def action_ZeroOperator(self, op):
        range_basis, source_basis = self.range_basis, self.source_basis
        if source_basis is not None and range_basis is not None:
            from pymor.operators.numpy import NumpyMatrixOperator
            return NumpyMatrixOperator(np.zeros((len(range_basis), len(source_basis))),
                                       name=op.name)
        else:
            new_source = NumpyVectorSpace(len(source_basis)) if source_basis is not None else op.source
            new_range = NumpyVectorSpace(len(range_basis)) if range_basis is not None else op.range
            return ZeroOperator(new_range, new_source, name=op.name)

    @match_class(ConstantOperator)
    def action_ConstantOperator(self, op):
        range_basis, source_basis = self.range_basis, self.source_basis
        if range_basis is not None:
            projected_value = NumpyVectorSpace.make_array(range_basis.inner(op.value).T)
        else:
            projected_value = op.value
        if source_basis is None:
            return ConstantOperator(projected_value, op.source, name=op.name)
        else:
            return ConstantOperator(projected_value, NumpyVectorSpace(len(source_basis)), name=op.name)

    @match_generic(lambda op: op.linear and not op.parametric, 'linear and not parametric')
    def action_apply_basis(self, op):
        range_basis, source_basis = self.range_basis, self.source_basis
        if source_basis is None:
            try:
                V = op.apply_adjoint(range_basis)
            except NotImplementedError:
                raise RuleNotMatchingError('apply_adjoint not implemented')
            if isinstance(op.source, NumpyVectorSpace):
                from pymor.operators.numpy import NumpyMatrixOperator
                return NumpyMatrixOperator(V.to_numpy(), source_id=op.source.id, name=op.name)
            else:
                from pymor.operators.constructions import VectorArrayOperator
                return VectorArrayOperator(V, adjoint=True, name=op.name)
        else:
            if range_basis is None:
                V = op.apply(source_basis)
                if isinstance(op.range, NumpyVectorSpace):
                    from pymor.operators.numpy import NumpyMatrixOperator
                    return NumpyMatrixOperator(V.to_numpy().T, range_id=op.range.id, name=op.name)
                else:
                    from pymor.operators.constructions import VectorArrayOperator
                    return VectorArrayOperator(V, adjoint=False, name=op.name)
            else:
                from pymor.operators.numpy import NumpyMatrixOperator
                return NumpyMatrixOperator(op.apply2(range_basis, source_basis), name=op.name)

    @match_class(ConcatenationOperator)
    def action_ConcatenationOperator(self, op):
        if len(op.operators) == 1:
            return self.apply(op.operators[0])

        range_basis, source_basis = self.range_basis, self.source_basis
        last, first = op.operators[0], op.operators[-1]

        if source_basis is not None and first.linear and not first.parametric:
            V = first.apply(source_basis)
            return project(op.with_(operators=op.operators[:-1]), range_basis, V)
        elif range_basis is not None and last.linear and not last.parametric:
            V = last.apply_adjoint(range_basis)
            return project(op.with_(operators=op.operators[1:]), V, source_basis)

        # the concatenation is too complicated to directly project efficiently
        # try to simplify the operators in the concatenation by expanding
        from pymor.algorithms.simplify import expand
        expanded_op = expand(op)
        if not isinstance(expanded_op, ConcatenationOperator):
            # expanding was successful
            return self.apply(expanded_op)

        # at least we can try to partially project the outer operators
        projected_first = project(first, None, source_basis)
        projected_last = project(last, range_basis, None)
        projected_op = ConcatenationOperator((projected_last,) + op.operators[1:-1] + (projected_first,), name=op.name)

        # special handling for concatenations with ConstantOperators
        # probably should be moved elsewhere
        if not projected_op.parametric and any(isinstance(o, ConstantOperator) for o in projected_op.operators):
            projected_op = ConstantOperator(projected_op.apply(projected_op.source.zeros()), projected_op.source)

        return projected_op

    @match_class(AdjointOperator)
    def action_AdjointOperator(self, op):
        range_basis, source_basis = self.range_basis, self.source_basis
        if range_basis is not None:
            if op.source_product:
                range_basis = op.source_product.apply_inverse(range_basis)

        if source_basis is not None and op.range_product:
            source_basis = op.range_product.apply(source_basis)

        operator = project(op.operator, source_basis, range_basis)
        range_product = op.range_product if source_basis is None else None
        source_product = op.source_product if range_basis is None else None
        return AdjointOperator(operator, source_product=source_product, range_product=range_product,
                               name=op.name)

    @match_class(EmpiricalInterpolatedOperator)
    def action_EmpiricalInterpolatedOperator(self, op):
        range_basis, source_basis = self.range_basis, self.source_basis
        if len(op.interpolation_dofs) == 0:
            return self.apply(ZeroOperator(op.range, op.source, op.name))
        elif not hasattr(op, 'restricted_operator') or source_basis is None:
            raise RuleNotMatchingError('Has no restricted operator or source_basis is None')
        if range_basis is not None:
            projected_collateral_basis = NumpyVectorSpace.make_array(op.collateral_basis.inner(range_basis))
        else:
            projected_collateral_basis = op.collateral_basis

        return ProjectedEmpiciralInterpolatedOperator(op.restricted_operator, op.interpolation_matrix,
                                                      NumpyVectorSpace.make_array(source_basis.dofs(op.source_dofs)),
                                                      projected_collateral_basis, op.triangular, None, op.name)

    @match_class(AffineOperator)
    def action_AffineOperator(self, op):
        return self.apply(op.affine_shift + op.linear_part)

    @match_class(LincombOperator)
    def action_LincombOperator(self, op):
        return self.replace_children(op).with_(solver_options=None)

    @match_class(SelectionOperator)
    def action_SelectionOperator(self, op):
        return self.replace_children(op)

    @match_class(BlockOperatorBase)
    def action_BlockOperatorBase(self, op):
        if op.blocked_range:
            if self.range_basis is not None:
                range_bases = self.range_basis._blocks
            else:
                range_bases = [None] * len(op.range.subspaces)
        else:
            range_bases = [self.range_basis]
        if op.blocked_source:
            if self.source_basis is not None:
                source_bases = self.source_basis._blocks
            else:
                source_bases = [None] * len(op.source.subspaces)
        else:
            source_bases = [self.source_basis]

        projected_ops = np.array([[project(op.blocks[i, j], rb, sb)
                                   for j, sb in enumerate(source_bases)]
                                  for i, rb in enumerate(range_bases)])
        if self.range_basis is None and op.blocked_range:
            return BlockColumnOperator(np.sum(projected_ops, axis=1))
        elif self.source_basis is None and op.blocked_source:
            return BlockRowOperator(np.sum(projected_ops, axis=0))
        else:
            return np.sum(projected_ops)


def project_to_subbasis(op, dim_range=None, dim_source=None):
    """Project already projected |Operator| to a subbasis.

    The purpose of this method is to further project an operator that has been
    obtained through :meth:`project` to subbases of the original projection bases, i.e. ::

        project_to_subbasis(project(op, r_basis, s_basis, prod), dim_range, dim_source)

    should be the same as ::

        project(op, r_basis[:dim_range], s_basis[:dim_source], prod)

    For a |NumpyMatrixOperator| this amounts to extracting the upper-left
    (dim_range, dim_source) corner of its matrix.

    The subbasis projection algorithm is specified in :class:`ProjectToSubbasisRules`.

    Parameters
    ----------
    dim_range
        Dimension of the range subbasis.
    dim_source
        Dimension of the source subbasis.

    Returns
    -------
    The projected |Operator|.
    """
    assert dim_source is None or (isinstance(op.source, NumpyVectorSpace) and dim_source <= op.source.dim)
    assert dim_range is None or (isinstance(op.range, NumpyVectorSpace) and dim_range <= op.range.dim)

    if dim_range is None and dim_source is None:
        return op

    return ProjectToSubbasisRules(dim_range, dim_source).apply(op)


class ProjectToSubbasisRules(RuleTable):
    """|RuleTable| for the :func:`project_to_subbasis` algorithm."""

    def __init__(self, dim_range, dim_source):
        super().__init__(use_caching=True)
        self.__auto_init(locals())

    @match_class(LincombOperator, SelectionOperator)
    def action_recurse(self, op):
        return self.replace_children(op)

    @match_class(NumpyMatrixOperator)
    def action_NumpyMatrixOperator(self, op):
        # copy instead of just slicing the matrix to ensure contiguous memory
        return NumpyMatrixOperator(op.matrix[:self.dim_range, :self.dim_source].copy(),
                                   solver_options=op.solver_options, name=op.name,
                                   source_id=op.source.id, range_id=op.range.id)

    @match_class(ConstantOperator)
    def action_ConstantOperator(self, op):
        dim_range, dim_source = self.dim_range, self.dim_source
        source = op.source if dim_source is None else NumpyVectorSpace(dim_source)
        value = op.value if dim_range is None else NumpyVectorSpace.make_array(op.value.to_numpy()[:, :dim_range])
        return ConstantOperator(value, source, name=op.name)

    @match_class(IdentityOperator)
    def action_IdentityOperator(self, op):
        dim_range, dim_source = self.dim_range, self.dim_source
        if dim_range != dim_source:
            raise RuleNotMatchingError('dim_range and dim_source must be equal.')
        space = op.source if dim_source is None else NumpyVectorSpace(dim_source)
        return IdentityOperator(space, name=op.name)

    @match_class(ZeroOperator)
    def action_ZeroOperator(self, op):
        dim_range, dim_source = self.dim_range, self.dim_source
        range_space = op.range if dim_range is None else NumpyVectorSpace(dim_range)
        source_space = op.source if dim_source is None else NumpyVectorSpace(dim_source)
        return ZeroOperator(range_space, source_space, name=op.name)

    @match_class(ProjectedEmpiciralInterpolatedOperator)
    def action_ProjectedEmpiciralInterpolatedOperator(self, op):
        if not isinstance(op.projected_collateral_basis.space, NumpyVectorSpace):
            raise NotImplementedError

        restricted_operator = op.restricted_operator

        old_pcb = op.projected_collateral_basis
        projected_collateral_basis = NumpyVectorSpace.make_array(old_pcb.to_numpy()[:, :self.dim_range])

        old_sbd = op.source_basis_dofs
        source_basis_dofs = NumpyVectorSpace.make_array(old_sbd.to_numpy()[:self.dim_source])

        return ProjectedEmpiciralInterpolatedOperator(restricted_operator, op.interpolation_matrix,
                                                      source_basis_dofs, projected_collateral_basis, op.triangular,
                                                      solver_options=op.solver_options, name=op.name)

    @match_class(VectorArrayOperator)
    def action_VectorArrayOperator(self, op):
        dim_range, dim_source = self.dim_range, self.dim_source
        if op.adjoint and dim_source is not None:
            raise RuleNotMatchingError
        if not op.adjoint and dim_range is not None:
            raise RuleNotMatchingError
        array = op.array[:(dim_range if op.adjoint else dim_source)]
        # use new_type to ensure this also works for child classes
        return op.with_(array=array, new_type=VectorArrayOperator)

    @match_class(BlockColumnOperator)
    def action_BlockColumnOperator(self, op):
        if self.dim_range is not None:
            raise RuleNotMatchingError
        blocks = [self.apply(b) for b in op.blocks.ravel()]
        return op.with_(blocks=blocks)

    @match_class(BlockRowOperator)
    def action_BlockRowOperator(self, op):
        if self.dim_source is not None:
            raise RuleNotMatchingError
        blocks = [self.apply(b) for b in op.blocks.ravel()]
        return op.with_(blocks=blocks)

    @match_class(ProjectedOperator)
    def action_ProjectedOperator(self, op):
        dim_range, dim_source = self.dim_range, self.dim_source
        source_basis = op.source_basis if dim_source is None \
            else op.source_basis[:dim_source]
        range_basis = op.range_basis if dim_range is None \
            else op.range_basis[:dim_range]
        return ProjectedOperator(op.operator, range_basis, source_basis, product=None,
                                 solver_options=op.solver_options)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from numbers import Number
import time

import numpy as np

from pymor.core.logger import getLogger
from pymor.models.basic import StationaryModel
from pymor.parallel.dummy import dummy_pool


def reduction_error_analysis(rom, fom, reductor, test_mus,
                             basis_sizes=0,
                             error_estimator=True, condition=False, error_norms=(),
                             error_norm_names=None,
                             error_estimator_norm_index=0, custom=(),
                             plot=False, plot_custom_logarithmic=True,
                             pool=dummy_pool):
    """Analyze the model reduction error.

    The maximum model reduction error is estimated by solving the reduced
    |Model| for given random |Parameters|.

    Parameters
    ----------
    rom
        The reduced |Model|.
    fom
        The high-dimensional |Model|.
    reductor
        The reductor which has created `rom`.
    test_mus
        List of |Parameters| to compute the errors for.
    basis_sizes
        Either a list of reduced basis dimensions to consider, or
        the number of dimensions (which are then selected equidistantly,
        always including the maximum reduced space dimension).
        The dimensions are input for the `dim`-Parameter of
        `reductor.reduce()`.
    error_estimator
        If `True` evaluate the error estimator of `rom`
        on the test |Parameters|.
    condition
        If `True`, compute the condition of the reduced system matrix
        for the given test |Parameters| (can only be specified if
        `rom` is an instance of |StationaryModel|
        and `rom.operator` is linear).
    error_norms
        List of norms in which to compute the model reduction error.
    error_norm_names
        Names of the norms given by `error_norms`. If `None`, the
        `name` attributes of the given norms are used.
    error_estimator_norm_index
        When `error_estimator` is `True` and `error_norms` are specified,
        this is the index of the norm in `error_norms` w.r.t. which
        to compute the effectivity of the error estimator.
    custom
        List of custom functions which are evaluated for each test
        |parameter values| and basis size. The functions must have
        the signature ::

            def custom_value(rom, fom, reductor, mu, dim):
                pass

    plot
        If `True`, generate a plot of the computed quantities w.r.t.
        the basis size.
    plot_custom_logarithmic
        If `True`, use a logarithmic y-axis to plot the computed custom
        values.
    pool
        If not `None`, the |WorkerPool| to use for parallelization.

    Returns
    -------
    Dict with the following fields:

        :mus:                       The test |Parameters| which have been considered.

        :basis_sizes:               The reduced basis dimensions which have been considered.

        :norms:                     |Array| of the norms of the high-dimensional solutions
                                    w.r.t. all given test |Parameters|, reduced basis
                                    dimensions and norms in `error_norms`.
                                    (Only present when `error_norms` has been specified.)

        :max_norms:                 Maxima of `norms` over the given test |Parameters|.

        :max_norm_mus:              |Parameters| corresponding to `max_norms`.

        :errors:                    |Array| of the norms of the model reduction errors
                                    w.r.t. all given test |Parameters|, reduced basis
                                    dimensions and norms in `error_norms`.
                                    (Only present when `error_norms` has been specified.)

        :max_errors:                Maxima of `errors` over the given test |Parameters|.

        :max_error_mus:             |Parameters| corresponding to `max_errors`.

        :rel_errors:                `errors` divided by `norms`.
                                    (Only present when `error_norms` has been specified.)

        :max_rel_errors:            Maxima of `rel_errors` over the given test |Parameters|.

        :max_rel_error_mus:         |Parameters| corresponding to `max_rel_errors`.

        :error_norm_names:          Names of the given `error_norms`.
                                    (Only present when `error_norms` has been specified.)

        :error_estimates:           |Array| of the model reduction error estimates
                                    w.r.t. all given test |Parameters| and reduced basis
                                    dimensions.
                                    (Only present when `error_estimator` is `True`.)

        :max_error_estimate:        Maxima of `error_estimates` over the given test |Parameters|.

        :max_error_estimate_mus:    |Parameters| corresponding to `max_error_estimates`.

        :effectivities:             `errors` divided by `error_estimates`.
                                    (Only present when `error_estimator` is `True` and `error_norms`
                                    has been specified.)

        :min_effectivities:         Minima of `effectivities` over the given test |Parameters|.

        :min_effectivity_mus:       |Parameters| corresponding to `min_effectivities`.

        :max_effectivities:         Maxima of `effectivities` over the given test |Parameters|.

        :max_effectivity_mus:       |Parameters| corresponding to `max_effectivities`.

        :errors:                    |Array| of the reduced system matrix conditions
                                    w.r.t. all given test |Parameters| and reduced basis
                                    dimensions.
                                    (Only present when `conditions` is `True`.)

        :max_conditions:            Maxima of `conditions` over the given test |Parameters|.

        :max_condition_mus:         |Parameters| corresponding to `max_conditions`.

        :custom_values:             |Array| of custom function evaluations
                                    w.r.t. all given test |Parameters|, reduced basis
                                    dimensions and functions in `custom`.
                                    (Only present when `custom` has been specified.)

        :max_custom_values:         Maxima of `custom_values` over the given test |Parameters|.

        :max_custom_values_mus:     |Parameters| corresponding to `max_custom_values`.

        :time:                      Time (in seconds) needed for the error analysis.

        :summary:                   String containing a summary of all computed quantities for
                                    the largest (last) considered basis size.

        :figure:                    The figure containing the generated plots.
                                    (Only present when `plot` is `True`.)
    """
    assert not error_norms or (fom and reductor)
    assert error_norm_names is None or len(error_norm_names) == len(error_norms)
    assert not condition \
        or isinstance(rom, StationaryModel) and rom.operator.linear

    logger = getLogger('pymor.algorithms.error')
    if pool is None or pool is dummy_pool:
        pool = dummy_pool
    else:
        logger.info(f'Using pool of {len(pool)} workers for error analysis')

    tic = time.perf_counter()

    if isinstance(basis_sizes, Number):
        if basis_sizes == 1:
            basis_sizes = [rom.solution_space.dim]
        else:
            if basis_sizes == 0:
                basis_sizes = rom.solution_space.dim + 1
            basis_sizes = min(rom.solution_space.dim + 1, basis_sizes)
            basis_sizes = np.linspace(0, rom.solution_space.dim, basis_sizes).astype(int)
    if error_norm_names is None:
        error_norm_names = tuple(norm.name for norm in error_norms)

    norms, error_estimates, errors, conditions, custom_values = \
        list(zip(*pool.map(_compute_errors, test_mus, fom=fom, reductor=reductor, error_estimator=error_estimator,
                           error_norms=error_norms, condition=condition, custom=custom, basis_sizes=basis_sizes)))
    print()

    result = {}
    result['mus'] = test_mus = np.array(test_mus)
    result['basis_sizes'] = basis_sizes

    summary = [('number of samples', str(len(test_mus)))]

    if error_norms:
        result['norms'] = norms = np.array(norms)
        result['max_norms'] = max_norms = np.max(norms, axis=0)
        result['max_norm_mus'] = max_norm_mus = test_mus[np.argmax(norms, axis=0)]
        result['errors'] = errors = np.array(errors)
        result['max_errors'] = max_errors = np.max(errors, axis=0)
        result['max_error_mus'] = max_error_mus = test_mus[np.argmax(errors, axis=0)]
        result['rel_errors'] = rel_errors = errors / norms[:, :, np.newaxis]
        result['max_rel_errors'] = np.max(rel_errors, axis=0)
        result['max_rel_error_mus'] = test_mus[np.argmax(rel_errors, axis=0)]
        for name, norm, norm_mu, error, error_mu in zip(error_norm_names,
                                                        max_norms, max_norm_mus,
                                                        max_errors[:, -1], max_error_mus[:, -1]):
            summary.append((f'maximum {name}-norm',
                            f'{norm:.7e} (mu = {error_mu})'))
            summary.append((f'maximum {name}-error',
                            f'{error:.7e} (mu = {error_mu})'))
        result['error_norm_names'] = error_norm_names

    if error_estimator:
        result['error_estimates'] = error_estimates = np.array(error_estimates)
        result['max_error_estimates'] = max_error_estimates = np.max(error_estimates, axis=0)
        result['max_error_estimate_mus'] = max_error_estimate_mus = test_mus[np.argmax(error_estimates, axis=0)]
        summary.append(('maximum estimated error',
                        f'{max_error_estimates[-1]:.7e} (mu = {max_error_estimate_mus[-1]})'))

    if error_estimator and error_norms:
        result['effectivities'] = effectivities = errors[:, error_estimator_norm_index, :] / error_estimates
        result['max_effectivities'] = max_effectivities = np.max(effectivities, axis=0)
        result['max_effectivity_mus'] = max_effectivity_mus = test_mus[np.argmax(effectivities, axis=0)]
        result['min_effectivities'] = min_effectivities = np.min(effectivities, axis=0)
        result['min_effectivity_mus'] = min_effectivity_mus = test_mus[np.argmin(effectivities, axis=0)]
        summary.append(('minimum error estimator effectivity',
                        f'{min_effectivities[-1]:.7e} (mu = {min_effectivity_mus[-1]})'))
        summary.append(('maximum error estimator effectivity',
                        f'{max_effectivities[-1]:.7e} (mu = {max_effectivity_mus[-1]})'))

    if condition:
        result['conditions'] = conditions = np.array(conditions)
        result['max_conditions'] = max_conditions = np.max(conditions, axis=0)
        result['max_condition_mus'] = max_condition_mus = test_mus[np.argmax(conditions, axis=0)]
        summary.append(('maximum system matrix condition',
                        f'{max_conditions[-1]:.7e} (mu = {max_condition_mus[-1]})'))

    if custom:
        result['custom_values'] = custom_values = np.array(custom_values)
        result['max_custom_values'] = max_custom_values = np.max(custom_values, axis=0)
        result['max_custom_values_mus'] = max_custom_values_mus = test_mus[np.argmax(custom_values, axis=0)]
        for i, (value, mu) in enumerate(zip(max_custom_values[:, -1], max_custom_values_mus[:, -1])):
            summary.append((f'maximum custom value {i}',
                            f'{value:.7e} (mu = {mu})'))

    toc = time.perf_counter()
    result['time'] = toc - tic
    summary.append(('elapsed time', str(toc - tic)))

    summary_fields, summary_values = list(zip(*summary))
    summary_field_width = np.max(list(map(len, summary_fields))) + 2
    summary_lines = [f'    {field+":":{summary_field_width}} {value}'
                     for field, value in zip(summary_fields, summary_values)]
    summary = 'Stochastic error estimation:\n' + '\n'.join(summary_lines)
    result['summary'] = summary

    if plot:
        import matplotlib.pyplot as plt
        fig = plt.figure()
        num_plots = (int(bool(error_norms) or error_estimator) + int(bool(error_norms) and error_estimator)
                     + int(condition) + int(bool(custom)))
        current_plot = 1

        if bool(error_norms) or error_estimator:
            ax = fig.add_subplot(1, num_plots, current_plot)
            legend = []
            if error_norms:
                for name, errors in zip(error_norm_names, max_errors):
                    ax.semilogy(basis_sizes, errors)
                    legend.append(name)
            if error_estimator:
                ax.semilogy(basis_sizes, max_error_estimates)
                legend.append('error estimator')
            ax.legend(legend)
            ax.set_title('maximum errors')
            current_plot += 1

        if bool(error_norms) and error_estimator:
            ax = fig.add_subplot(1, num_plots, current_plot)
            ax.semilogy(basis_sizes, min_effectivities)
            ax.semilogy(basis_sizes, max_effectivities)
            ax.legend(('min', 'max'))
            ax.set_title('error estimator effectivities')
            current_plot += 1

        if condition:
            ax = fig.add_subplot(1, num_plots, current_plot)
            ax.semilogy(basis_sizes, max_conditions)
            ax.set_title('maximum condition')
            current_plot += 1

        if custom:
            ax = fig.add_subplot(1, num_plots, current_plot)
            legend = []
            for i, values in enumerate(custom_values):
                if plot_custom_logarithmic:
                    ax.semilogy(basis_sizes, values)
                else:
                    ax.plot(basis_sizes, values)
                legend.append('value ' + str(i))
            ax.legend(legend)
            ax.set_title('maximum custom values')
            current_plot += 1

        result['figure'] = fig

    return result


def _compute_errors(mu, fom, reductor, error_estimator, error_norms, condition, custom, basis_sizes):
    import sys

    print('.', end='')
    sys.stdout.flush()

    error_estimates = np.empty(len(basis_sizes)) if error_estimator else None
    norms = np.empty(len(error_norms))
    errors = np.empty((len(error_norms), len(basis_sizes)))
    conditions = np.empty(len(basis_sizes)) if condition else None
    custom_values = np.empty((len(custom), len(basis_sizes)))

    if fom:
        logging_disabled = fom.logging_disabled
        fom.disable_logging()
        U = fom.solve(mu)
        fom.disable_logging(logging_disabled)
        for i_norm, norm in enumerate(error_norms):
            n = norm(U)
            n = n[0] if hasattr(n, '__len__') else n
            norms[i_norm] = n

    for i_N, N in enumerate(basis_sizes):
        rom = reductor.reduce(dims={k: N for k in reductor.bases})
        result = rom.compute(solution=True, solution_error_estimate=error_estimator, mu=mu)
        u = result['solution']
        if error_estimator:
            e = result['solution_error_estimate']
            e = e[0] if hasattr(e, '__len__') else e
            error_estimates[i_N] = e
        if fom and reductor:
            URB = reductor.reconstruct(u)
            for i_norm, norm in enumerate(error_norms):
                e = norm(U - URB)
                e = e[0] if hasattr(e, '__len__') else e
                errors[i_norm, i_N] = e
        if condition:
            conditions[i_N] = np.linalg.cond(rom.operator.assemble(mu).matrix) if N > 0 else 0.
        for i_custom, cust in enumerate(custom):
            c = cust(rom=rom,
                     fom=fom,
                     reductor=reductor,
                     mu=mu,
                     dim=N)
            c = c[0] if hasattr(c, '__len__') else c
            custom_values[i_custom, i_N] = c

    return norms, error_estimates, errors, conditions, custom_values



#!/usr/bin/env python
# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
import matplotlib.pyplot as plt
from typer import Argument, run


from pymor.analyticalproblems.domaindescriptions import RectDomain
from pymor.analyticalproblems.elliptic import StationaryProblem
from pymor.analyticalproblems.functions import ConstantFunction, ExpressionFunction
from pymor.analyticalproblems.instationary import InstationaryProblem
from pymor.core.config import config
from pymor.core.logger import set_log_levels
from pymor.discretizers.builtin import discretize_instationary_cg
from pymor.models.iosys import LTIModel
from pymor.reductors.bt import BTReductor, LQGBTReductor, BRBTReductor
from pymor.reductors.h2 import IRKAReductor, TSIAReductor, OneSidedIRKAReductor
from pymor.reductors.mt import MTReductor


def fom_properties(fom, w):
    """Show properties of the full-order model.

    Parameters
    ----------
    fom
        The full-order `InputStateOutputModel`.
    w
        Array of frequencies.
    """
    from pymor.models.iosys import TransferFunction
    if not isinstance(fom, TransferFunction):
        print(f'order of the model = {fom.order}')
    print(f'number of inputs   = {fom.dim_input}')
    print(f'number of outputs  = {fom.dim_output}')

    # System norms
    print(f'FOM H_2-norm:    {fom.h2_norm():e}')
    if not isinstance(fom, TransferFunction):
        if config.HAVE_SLYCOT:
            print(f'FOM H_inf-norm:  {fom.hinf_norm():e}')
        else:
            print('Skipped H_inf-norm calculation due to missing slycot.')
        print(f'FOM Hankel-norm: {fom.hankel_norm():e}')

    # System poles
    if not isinstance(fom, TransferFunction):
        poles = fom.poles()
        fig, ax = plt.subplots()
        ax.plot(poles.real, poles.imag, '.')
        ax.set_title('System poles')

    # Bode plot of the full model
    fig, ax = plt.subplots(2 * fom.dim_output, fom.dim_input, squeeze=False)
    fom.bode_plot(w, ax=ax)
    fig.suptitle('Bode plot of the full model')
    plt.show()


def run_mor_method(lti, w, reductor, reductor_short_name, r, **reduce_kwargs):
    """Run a model order reduction method.

    Parameters
    ----------
    lti
        The full-order `InputStateOutputModel`.
    w
        Array of frequencies.
    reductor
        The reductor object.
    reductor_short_name
        A short name for the reductor.
    r
        The order of the reduced-order model.
    reduce_kwargs
        Optional keyword arguments for the reduce method.
    """
    # Reduction
    rom = reductor.reduce(r, **reduce_kwargs)
    err = lti - rom

    # Errors
    from pymor.models.iosys import TransferFunction
    if not isinstance(lti, TransferFunction):
        print(f'{reductor_short_name} relative H_2-error:    {err.h2_norm() / lti.h2_norm():e}')
        if config.HAVE_SLYCOT:
            print(f'{reductor_short_name} relative H_inf-error:  {err.hinf_norm() / lti.hinf_norm():e}')
        else:
            print('Skipped H_inf-norm calculation due to missing slycot.')
        print(f'{reductor_short_name} relative Hankel-error: {err.hankel_norm() / lti.hankel_norm():e}')
    elif isinstance(rom, LTIModel):
        error = np.sqrt(lti.h2_norm()**2 - 2 * lti.h2_inner(rom).real + rom.h2_norm()**2)
        print(f'{reductor_short_name} relative H_2-error:    {error / lti.h2_norm():e}')
    else:
        print(f'{reductor_short_name} relative H_2-error:    {err.h2_norm() / lti.h2_norm():e}')

    # Poles of the reduced-order model
    poles_rom = rom.poles()
    fig, ax = plt.subplots()
    ax.plot(poles_rom.real, poles_rom.imag, '.')
    ax.set_title(f"{reductor_short_name} reduced model's poles")

    # Bode plot of the full and reduced model
    fig, ax = plt.subplots(2 * lti.dim_output, lti.dim_input, squeeze=False)
    lti.bode_plot(w, ax=ax)
    rom.bode_plot(w, ax=ax, linestyle='dashed')
    fig.suptitle(f'Bode plot of the full and {reductor_short_name} reduced model')

    # Magnitude plot of the error system
    fig, ax = plt.subplots()
    err.mag_plot(w, ax=ax)
    ax.set_title(f'Magnitude plot of the {reductor_short_name} error system')
    plt.show()


def main(
        diameter: float = Argument(0.1, help='Diameter option for the domain discretizer.'),
        r: int = Argument(5, help='Order of the ROMs.'),
):
    r"""2D heat equation demo.

    Discretization of the PDE:

    .. math::
        \begin{align*}
            \partial_t z(x, y, t) &= \Delta z(x, y, t),      & 0 < x, y < 1,\ t > 0 \\
            -\nabla z(0, y, t) \cdot n &= z(0, y, t) - u(t), & 0 < y < 1, t > 0 \\
            -\nabla z(1, y, t) \cdot n &= z(1, y, t),        & 0 < y < 1, t > 0 \\
            -\nabla z(0, x, t) \cdot n &= z(0, x, t),        & 0 < x < 1, t > 0 \\
            -\nabla z(1, x, t) \cdot n &= z(1, x, t),        & 0 < x < 1, t > 0 \\
            z(x, y, 0) &= 0                                  & 0 < x, y < 1 \\
            y(t) &= \int_0^1 z(1, y, t) dy,                  & t > 0
        \end{align*}

    where :math:`u(t)` is the input and :math:`y(t)` is the output.
    """
    set_log_levels({'pymor.algorithms.gram_schmidt.gram_schmidt': 'WARNING'})

    p = InstationaryProblem(
        StationaryProblem(
            domain=RectDomain([[0., 0.], [1., 1.]], left='robin', right='robin', top='robin', bottom='robin'),
            diffusion=ConstantFunction(1., 2),
            robin_data=(ConstantFunction(1., 2), ExpressionFunction('(x[0] < 1e-10) * 1.', 2)),
            outputs=[('l2_boundary', ExpressionFunction('(x[0] > (1 - 1e-10)) * 1.', 2))]
        ),
        ConstantFunction(0., 2),
        T=1.
    )

    fom, _ = discretize_instationary_cg(p, diameter=diameter, nt=100)

    fom.visualize(fom.solve())

    # LTI system
    lti = fom.to_lti()

    # System properties
    w = np.logspace(-1, 3, 100)
    fom_properties(lti, w)

    # Hankel singular values
    hsv = lti.hsv()
    fig, ax = plt.subplots()
    ax.semilogy(range(1, len(hsv) + 1), hsv, '.-')
    ax.set_title('Hankel singular values')
    plt.show()

    # Model order reduction
    run_mor_method(lti, w, BTReductor(lti), 'BT', r, tol=1e-5)
    run_mor_method(lti, w, LQGBTReductor(lti), 'LQGBT', r, tol=1e-5)
    run_mor_method(lti, w, BRBTReductor(lti), 'BRBT', r, tol=1e-5)
    run_mor_method(lti, w, IRKAReductor(lti), 'IRKA', r)
    run_mor_method(lti, w, IRKAReductor(lti), 'IRKA (with Arnoldi)', r, projection='arnoldi')
    run_mor_method(lti, w, TSIAReductor(lti), 'TSIA', r)
    run_mor_method(lti, w, OneSidedIRKAReductor(lti, 'V'), 'OS-IRKA', r)
    run_mor_method(lti, w, MTReductor(lti), 'MT', r)


if __name__ == '__main__':
    run(main)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pathlib import Path

import numpy as np
from scipy.io import loadmat, mmread, mmwrite, savemat
from scipy.sparse import issparse

from pymor.core.logger import getLogger


def _loadmat(path, key=None):
    try:
        data = loadmat(path, mat_dtype=True)
    except Exception as e:
        raise IOError(e)

    if key:
        try:
            return data[key]
        except KeyError:
            raise IOError(f'"{key}" not found in MATLAB file {path}')

    data = [v for v in data.values() if isinstance(v, np.ndarray) or issparse(v)]

    if len(data) == 0:
        raise IOError(f'No matrix data contained in MATLAB file {path}')
    elif len(data) > 1:
        raise IOError(f'More than one matrix object stored in MATLAB file {path}')
    else:
        return data[0]


def _savemat(path, matrix, key=None):
    if key is None:
        raise IOError('"key" must be specified for MATLAB file')
    try:
        savemat(path, {key: matrix})
    except Exception as e:
        raise IOError(e)


def _mmread(path, key=None):
    if key:
        raise IOError('Cannot specify "key" for Matrix Market file')
    try:
        matrix = mmread(path)
        if issparse(matrix):
            matrix = matrix.tocsc()
        return matrix
    except Exception as e:
        raise IOError(e)


def _mmwrite(path, matrix, key=None):
    if key:
        raise IOError('Cannot specify "key" for Matrix Market file')
    try:
        if path.suffix != '.gz':
            open_file = open
        else:
            import gzip
            open_file = gzip.open
        with open_file(path, 'wb') as f:
            # when mmwrite is given a string, it will append '.mtx'
            mmwrite(f, matrix)
    except Exception as e:
        raise IOError(e)


def _load(path, key=None):
    try:
        data = np.load(path)
    except Exception as e:
        raise IOError(e)
    if isinstance(data, (dict, np.lib.npyio.NpzFile)):
        if key:
            try:
                matrix = data[key]
            except KeyError:
                raise IOError(f'"{key}" not found in NPY file {path}')
        elif len(data) == 0:
            raise IOError(f'No data contained in NPY file {path}')
        elif len(data) > 1:
            raise IOError(f'More than one object stored in NPY file {path} for key {key}')
        else:
            matrix = next(iter(data.values()))
    else:
        matrix = data
    if not isinstance(matrix, np.ndarray) and not issparse(matrix):
        raise IOError(f'Loaded data is not a matrix in NPY file {path}')
    return matrix


def _save(path, matrix, key=None):
    if key:
        raise IOError('Cannot specify "key" for NPY file')
    try:
        np.save(path, matrix)
    except Exception as e:
        raise IOError(e)


def _savez(path, matrix, key=None):
    try:
        if key is None:
            np.savez(path, matrix)
        else:
            np.savez(path, **{key: matrix})
    except Exception as e:
        raise IOError(e)


def _loadtxt(path, key=None):
    if key:
        raise IOError('Cannot specify "key" for TXT file')
    try:
        return np.loadtxt(path)
    except Exception as e:
        raise IOError(e)


def _savetxt(path, matrix, key=None):
    if key:
        raise IOError('Cannot specify "key" for TXT file')
    try:
        return np.savetxt(path, matrix)
    except Exception as e:
        raise IOError(e)


def _get_file_extension(path):
    suffix_count = len(path.suffixes)
    if suffix_count and len(path.suffixes[-1]) == 4:
        extension = path.suffixes[-1].lower()
    elif path.suffixes[-1].lower() == '.gz' and suffix_count >= 2 and len(path.suffixes[-2]) == 4:
        extension = ''.join(path.suffixes[-2:]).lower()
    else:
        extension = ''
    return extension


def load_matrix(path, key=None):
    """Load matrix from file.

    Parameters
    ----------
    path
        Path to the file (`str` or `pathlib.Path`).
    key
        Key of the matrix (only for NPY, NPZ, and MATLAB files).

    Returns
    -------
    matrix
        |NumPy array| of |SciPy spmatrix|.

    Raises
    ------
    IOError
        If loading fails.
    """
    logger = getLogger('pymor.tools.io.load_matrix')
    logger.info('Loading matrix from file %s', path)

    # convert if path is str
    path = Path(path)
    extension = _get_file_extension(path)

    file_format_map = {
        '.mat': ('MATLAB', _loadmat),
        '.mtx': ('Matrix Market', _mmread),
        '.mtz.gz': ('Matrix Market', _mmread),
        '.npy': ('NPY/NPZ', _load),
        '.npz': ('NPY/NPZ', _load),
        '.txt': ('Text', _loadtxt),
    }

    if extension in file_format_map:
        file_type, loader = file_format_map[extension]
        logger.info(file_type + ' file detected.')
        return loader(path, key)

    logger.warning('Could not detect file format. Trying all loaders ...')

    loaders = [_loadmat, _mmread, _loadtxt, _load]
    for loader in loaders:
        try:
            return loader(path, key)
        except IOError:
            pass

    raise IOError(f'Could not load file {path} (key = {key})')


def save_matrix(path, matrix, key=None):
    """Save matrix to file.

    Parameters
    ----------
    path
        Path to the file (`str` or `pathlib.Path`).
    matrix
        Matrix to save.
    key
        Key of the matrix (only for NPY, NPZ, and MATLAB files).

    Raises
    ------
    IOError
        If saving fails.
    """
    logger = getLogger('pymor.tools.io.save_matrix')
    logger.info('Saving matrix to file %s', path)

    # convert if path is str
    path = Path(path)
    extension = _get_file_extension(path)

    file_format_map = {
        '.mat': ('MATLAB', _savemat),
        '.mtx': ('Matrix Market', _mmwrite),
        '.mtz.gz': ('Matrix Market', _mmwrite),
        '.npy': ('NPY', _save),
        '.npz': ('NPZ', _savez),
        '.txt': ('Text', _savetxt),
    }

    if extension in file_format_map:
        file_type, saver = file_format_map[extension]
        logger.info(file_type + ' file detected.')
        return saver(path, matrix, key)

    raise IOError(f'Unknown extension "{extension}"')



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import functools
import types
import inspect
import warnings


class Deprecated:
    """Decorator for marking functions as deprecated.

    It will result in a warning being emitted when the function is used.
    """

    def __init__(self, alt='no alternative given'):
        if hasattr(alt, '__qualname__'):
            alt = f'{alt.__module__}.{alt.__qualname__}'
        self._alt = alt

    def __call__(self, func):
        func.decorated = self

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            func_name = f'{func.__module__}.{func.__qualname__}'
            frame = inspect.currentframe().f_back
            msg = f'DeprecationWarning. Call to deprecated function {func_name}  in ' \
                  f'{frame.f_code.co_filename}:{frame.f_code.co_firstlineno}\n' \
                  f'Use {self._alt} instead'
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return new_func

    def __get__(self, obj, ownerClass=None):
        """Return a wrapper that binds self as a method of obj (!)"""
        self.obj = obj
        return types.MethodType(self, obj)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""This module contains algorithms for the empirical interpolation of |Operators|.

The main work for generating the necessary interpolation data is handled by
the :func:`ei_greedy` method. The objects returned by this method can be used
to instantiate an |EmpiricalInterpolatedOperator|.

As a convenience, the :func:`interpolate_operators` method allows to perform
the empirical interpolation of the |Operators| of a given model with
a single function call.
"""

import numpy as np
from scipy.linalg import solve

from pymor.core.logger import getLogger
from pymor.algorithms.pod import pod as pod_alg
from pymor.analyticalproblems.functions import Function, EmpiricalInterpolatedFunction
from pymor.operators.ei import EmpiricalInterpolatedOperator
from pymor.parallel.dummy import dummy_pool
from pymor.parallel.interface import RemoteObject
from pymor.parallel.manager import RemoteObjectManager
from pymor.vectorarrays.interface import VectorArray
from pymor.vectorarrays.numpy import NumpyVectorSpace


def ei_greedy(U, error_norm=None, atol=None, rtol=None, max_interpolation_dofs=None,
              nodal_basis=False, copy=True, pool=dummy_pool):
    """Generate data for empirical interpolation using EI-Greedy algorithm.

    Given a |VectorArray| `U`, this method generates a collateral basis and
    interpolation DOFs for empirical interpolation of the vectors contained in `U`.
    The returned objects can be used to instantiate an |EmpiricalInterpolatedOperator|
    (with `triangular=True`).

    The interpolation data is generated by a greedy search algorithm, where in each
    loop iteration the worst approximated vector in `U` is added to the collateral basis.

    Parameters
    ----------
    U
        A |VectorArray| of vectors to interpolate.
    error_norm
        Norm w.r.t. which to calculate the interpolation error. If `None`, the Euclidean norm
        is used. If `'sup'`, the sup-norm of the dofs is used.
    atol
        Stop the greedy search if the largest approximation error is below this threshold.
    rtol
        Stop the greedy search if the largest relative approximation error is below this threshold.
    max_interpolation_dofs
        Stop the greedy search if the number of interpolation DOF (= dimension of the collateral
        basis) reaches this value.
    nodal_basis
        If `True`, a nodal interpolation basis is constructed. Note that nodal bases are
        not hierarchical. Their construction involves the inversion of the associated
        interpolation matrix, which might lead to decreased numerical accuracy.
    copy
        If `False`, `U` will be modified during executing of the algorithm.
    pool
        If not `None`, the |WorkerPool| to use for parallelization.

    Returns
    -------
    interpolation_dofs
        |NumPy array| of the DOFs at which the vectors are evaluated.
    collateral_basis
        |VectorArray| containing the generated collateral basis.
    data
        Dict containing the following fields:

            :errors:                Sequence of maximum approximation errors during
                                    greedy search.
            :triangularity_errors:  Sequence of maximum absolute values of interoplation
                                    matrix coefficients in the upper triangle (should
                                    be near zero).
            :coefficients:          |NumPy array| of coefficients such that `collateral_basis`
                                    is given by `U.lincomb(coefficients)`.
            :interpolation_matrix:  The interpolation matrix, i.e., the evaluation of
                                    `collateral_basis` at `interpolation_dofs`.
    """
    assert not isinstance(error_norm, str) or error_norm == 'sup'
    if pool:  # dispatch to parallel implementation
        assert isinstance(U, (VectorArray, RemoteObject))
        with RemoteObjectManager() as rom:
            if isinstance(U, VectorArray):
                U = rom.manage(pool.scatter_array(U))
            return _parallel_ei_greedy(U, error_norm=error_norm, atol=atol, rtol=rtol,
                                       max_interpolation_dofs=max_interpolation_dofs, copy=copy, pool=pool)

    assert isinstance(U, VectorArray)

    logger = getLogger('pymor.algorithms.ei.ei_greedy')
    logger.info('Generating Interpolation Data ...')

    interpolation_dofs = np.zeros((0,), dtype=np.int32)
    collateral_basis = U.empty()
    K = np.eye(len(U))  # matrix s.t. U = U_initial.lincomb(K)
    coefficients = np.zeros((0, len(U)))
    max_errs = []
    triangularity_errs = []

    if copy:
        U = U.copy()

    ERR = U

    errs = ERR.norm() if error_norm is None else ERR.sup_norm() if error_norm == 'sup' else error_norm(ERR)
    max_err_ind = np.argmax(errs)
    initial_max_err = max_err = errs[max_err_ind]

    # main loop
    while True:
        if max_interpolation_dofs is not None and len(interpolation_dofs) >= max_interpolation_dofs:
            logger.info('Maximum number of interpolation DOFs reached. Stopping extension loop.')
            logger.info(f'Final maximum interpolation error with '
                        f'{len(interpolation_dofs)} interpolation DOFs: {max_err}')
            break

        logger.info(f'Maximum interpolation error with '
                    f'{len(interpolation_dofs)} interpolation DOFs: {max_err}')

        if atol is not None and max_err <= atol:
            logger.info('Absolute error tolerance reached! Stopping extension loop.')
            break

        if rtol is not None and max_err / initial_max_err <= rtol:
            logger.info('Relative error tolerance reached! Stopping extension loop.')
            break

        # compute new interpolation dof and collateral basis vector
        new_vec = U[max_err_ind].copy()
        new_dof = new_vec.amax()[0][0]
        if new_dof in interpolation_dofs:
            logger.info(f'DOF {new_dof} selected twice for interplation! Stopping extension loop.')
            break
        new_dof_value = new_vec.dofs([new_dof])[0, 0]
        if new_dof_value == 0.:
            logger.info(f'DOF {new_dof} selected for interpolation has zero maximum error! Stopping extension loop.')
            break
        new_vec *= 1 / new_dof_value
        interpolation_dofs = np.hstack((interpolation_dofs, new_dof))
        collateral_basis.append(new_vec)
        coefficients = np.vstack([coefficients, K[max_err_ind] / new_dof_value])
        max_errs.append(max_err)

        # update U and ERR
        new_dof_values = U.dofs([new_dof])
        U.axpy(-new_dof_values[:, 0], new_vec)
        K -= (K[max_err_ind] / new_dof_value) * new_dof_values
        errs = ERR.norm() if error_norm is None else ERR.sup_norm() if error_norm == 'sup' else error_norm(ERR)
        max_err_ind = np.argmax(errs)
        max_err = errs[max_err_ind]

    interpolation_matrix = collateral_basis.dofs(interpolation_dofs).T
    triangularity_errors = np.abs(interpolation_matrix - np.tril(interpolation_matrix))
    for d in range(1, len(interpolation_matrix) + 1):
        triangularity_errs.append(np.max(triangularity_errors[:d, :d]))

    if len(triangularity_errs) > 0:
        logger.info(f'Interpolation matrix is not lower triangular with maximum error of {triangularity_errs[-1]}')

    if nodal_basis:
        logger.info('Building nodal basis.')
        inv_interpolation_matrix = np.linalg.inv(interpolation_matrix)
        collateral_basis = collateral_basis.lincomb(inv_interpolation_matrix.T)
        coefficients = inv_interpolation_matrix.T @ coefficients
        interpolation_matrix = np.eye(len(collateral_basis))

    data = {'errors': max_errs, 'triangularity_errors': triangularity_errs,
            'coefficients': coefficients, 'interpolation_matrix': interpolation_matrix}

    return interpolation_dofs, collateral_basis, data


def deim(U, modes=None, pod=True, atol=None, rtol=None, product=None, pod_options={}):
    """Generate data for empirical interpolation using DEIM algorithm.

    Given a |VectorArray| `U`, this method generates a collateral basis and
    interpolation DOFs for empirical interpolation of the vectors contained in `U`.
    The returned objects can be used to instantiate an |EmpiricalInterpolatedOperator|
    (with `triangular=False`).

    The collateral basis is determined by the first :func:`~pymor.algorithms.pod.pod` modes of `U`.

    Parameters
    ----------
    U
        A |VectorArray| of vectors to interpolate.
    modes
        Dimension of the collateral basis i.e. number of POD modes of the vectors in `U`.
    pod
        If `True`, perform a POD of `U` to obtain the collateral basis. If `False`, `U`
        is used as collateral basis.
    atol
        Absolute POD tolerance.
    rtol
        Relative POD tolerance.
    product
        Inner product |Operator| used for the POD.
    pod_options
        Dictionary of additional options to pass to the :func:`~pymor.algorithms.pod.pod` algorithm.

    Returns
    -------
    interpolation_dofs
        |NumPy array| of the DOFs at which the vectors are interpolated.
    collateral_basis
        |VectorArray| containing the generated collateral basis.
    data
        Dict containing the following fields:

            :svals: POD singular values.
    """
    assert isinstance(U, VectorArray)

    logger = getLogger('pymor.algorithms.ei.deim')
    logger.info('Generating Interpolation Data ...')

    data = {}

    if pod:
        collateral_basis, svals = pod_alg(U, modes=modes, atol=atol, rtol=rtol, product=product, **pod_options)
        data['svals'] = svals
    else:
        collateral_basis = U

    interpolation_dofs = np.zeros((0,), dtype=np.int32)
    interpolation_matrix = np.zeros((0, 0))

    for i in range(len(collateral_basis)):
        logger.info(f'Choosing interpolation point for basis vector {i}.')

        if len(interpolation_dofs) > 0:
            coefficients = solve(interpolation_matrix,
                                 collateral_basis[i].dofs(interpolation_dofs).T).T
            U_interpolated = collateral_basis[:len(interpolation_dofs)].lincomb(coefficients)
            ERR = collateral_basis[i] - U_interpolated
        else:
            ERR = collateral_basis[i]

        # compute new interpolation dof and collateral basis vector
        new_dof = ERR.amax()[0][0]

        if new_dof in interpolation_dofs:
            logger.info(f'DOF {new_dof} selected twice for interplation! Stopping extension loop.')
            break

        interpolation_dofs = np.hstack((interpolation_dofs, new_dof))
        interpolation_matrix = collateral_basis[:len(interpolation_dofs)].dofs(interpolation_dofs).T

    if len(interpolation_dofs) < len(collateral_basis):
        del collateral_basis[len(interpolation_dofs):len(collateral_basis)]

    logger.info('Finished.')

    return interpolation_dofs, collateral_basis, data


def interpolate_operators(fom, operator_names, parameter_sample, error_norm=None,
                          product=None, atol=None, rtol=None, max_interpolation_dofs=None,
                          pod_options={}, alg='ei_greedy', pool=dummy_pool):
    """Empirical operator interpolation using the EI-Greedy/DEIM algorithm.

    This is a convenience method to facilitate the use of :func:`ei_greedy` or :func:`deim`.
    Given a |Model|, names of |Operators|, and a sample of |Parameters|, first
    the operators are evaluated on the solution snapshots of the model for the
    provided parameters. These evaluations are then used as input for
    :func:`ei_greedy`/:func:`deim`.  Finally the resulting interpolation data is used to
    create |EmpiricalInterpolatedOperators| and a new model with the interpolated
    operators is returned.

    Note that this implementation creates *one* common collateral basis for all specified
    operators, which might not be what you want.

    Parameters
    ----------
    fom
        The |Model| whose |Operators| will be interpolated.
    operator_names
        List of keys in the `operators` dict of the model. The corresponding
        |Operators| will be interpolated.
    parameter_sample
        A list of |Parameters| for which solution snapshots are calculated.
    error_norm
        See :func:`ei_greedy`.
        Has no effect if `alg == 'deim'`.
    product
        Inner product for POD computation in :func:`deim`.
        Has no effect if `alg == 'ei_greedy'`.
    atol
        See :func:`ei_greedy`.
    rtol
        See :func:`ei_greedy`.
    max_interpolation_dofs
        See :func:`ei_greedy`.
    pod_options
        Further options for :func:`~pymor.algorithms.pod.pod` algorithm.
        Has no effect if `alg == 'ei_greedy'`.
    alg
        Either `ei_greedy` or `deim`.
    pool
        If not `None`, the |WorkerPool| to use for parallelization.

    Returns
    -------
    eim
        |Model| with |Operators| given by `operator_names` replaced by
        |EmpiricalInterpolatedOperators|.
    data
        Dict containing the following fields:

            :dofs:     |NumPy array| of the DOFs at which the |Operators| have to be evaluated.
            :basis:    |VectorArray| containing the generated collateral basis.

        In addition, `data` contains the fields of the `data` `dict` returned by
        :func:`ei_greedy`/:func:`deim`.
    """
    assert alg in ('ei_greedy', 'deim')
    logger = getLogger('pymor.algorithms.ei.interpolate_operators')
    with RemoteObjectManager() as rom:
        operators = [getattr(fom, operator_name) for operator_name in operator_names]
        with logger.block('Computing operator evaluations on solution snapshots ...'):
            if pool:
                logger.info(f'Using pool of {len(pool)} workers for parallel evaluation')
                evaluations = rom.manage(pool.push(fom.solution_space.empty()))
                pool.map(_interpolate_operators_build_evaluations, parameter_sample,
                         fom=fom, operators=operators, evaluations=evaluations)
            else:
                evaluations = operators[0].range.empty()
                for mu in parameter_sample:
                    U = fom.solve(mu)
                    for op in operators:
                        evaluations.append(op.apply(U, mu=mu))

        if alg == 'ei_greedy':
            with logger.block('Performing EI-Greedy:'):
                dofs, basis, data = ei_greedy(evaluations, error_norm, atol=atol, rtol=rtol,
                                              max_interpolation_dofs=max_interpolation_dofs,
                                              copy=False, pool=pool)
        elif alg == 'deim':
            if alg == 'deim' and pool is not dummy_pool:
                logger.warning('DEIM algorithm not parallel. Collecting operator evaluations.')
                evaluations = pool.apply(_identity, x=evaluations)
                evs = evaluations[0]
                for e in evaluations[1:]:
                    evs.append(e, remove_from_other=True)
                evaluations = evs
            with logger.block('Executing DEIM algorithm:'):
                dofs, basis, data = deim(evaluations, modes=max_interpolation_dofs,
                                         atol=atol, rtol=rtol, pod_options=pod_options, product=product)
        else:
            assert False

    ei_operators = {name: EmpiricalInterpolatedOperator(operator, dofs, basis, triangular=(alg == 'ei_greedy'))
                    for name, operator in zip(operator_names, operators)}
    eim = fom.with_(name=f'{fom.name}_ei', **ei_operators)

    data.update({'dofs': dofs, 'basis': basis})
    return eim, data


def interpolate_function(function, parameter_sample, evaluation_points,
                         atol=None, rtol=None, max_interpolation_dofs=None):
    """Parameter separable approximation of a |Function| using Empiricial Interpolation.

    This method computes a parameter separated |LincombFunction| approximating
    the input |Function| using Empirical Interpolation :cite`BMNP04`.
    The actual EI Greedy algorithm is contained in :func:`ei_greedy`. This function
    acts as a convenience wrapper, which computes the training data and
    constructs an :class:`~pymor.analyticalproblems.functions.EmpiricalInterpolatedFunction`
    from the data returned by :func:`ei_greedy`.

    .. note::
        If possible, choose `evaluation_points` identical to the coordinates at which
        the interpolated function is going to be evaluated. Otherwise `function` will
        have to be re-evaluated at all new evaluation points for all |parameter values|
        given by `parameter_sample`.


    Parameters
    ----------
    function
        The function to interpolate.
    parameter_sample
        A list of |Parameters| for which `function` is evaluated to generate the
        training data.
    evaluation_points
        |NumPy array| of coordinates at which `function` should be evaluated to
        generate the training data.
    atol
        See :func:`ei_greedy`.
    rtol
        See :func:`ei_greedy`.
    max_interpolation_dofs
        See :func:`ei_greedy`.

    Returns
    -------
    ei_function
        The :class:`~pymor.analyticalproblems.functions.EmpiricalInterpolatedFunction` giving
        the parameter separable approximation of `function`.
    data
        `dict` of additional data as returned by :func:`ei_greedy`.
    """
    assert isinstance(function, Function)
    assert isinstance(evaluation_points, np.ndarray) and evaluation_points.ndim == 2 and \
        evaluation_points.shape[1] == function.dim_domain

    snapshot_data = NumpyVectorSpace.from_numpy(
        np.array([function(evaluation_points, mu=mu) for mu in parameter_sample])
    )

    dofs, basis, ei_data = ei_greedy(snapshot_data, error_norm='sup',
                                     atol=atol, rtol=rtol, max_interpolation_dofs=max_interpolation_dofs)

    ei_function = EmpiricalInterpolatedFunction(
        function, evaluation_points[dofs], ei_data['interpolation_matrix'], True,
        parameter_sample, ei_data['coefficients'],
        evaluation_points=evaluation_points, basis_evaluations=basis.to_numpy()
    )

    return ei_function, ei_data


def _interpolate_operators_build_evaluations(mu, fom=None, operators=None, evaluations=None):
    U = fom.solve(mu)
    for op in operators:
        evaluations.append(op.apply(U, mu=mu))


def _parallel_ei_greedy(U, pool, error_norm=None, atol=None, rtol=None, max_interpolation_dofs=None,
                        nodal_basis=False, copy=True):

    assert isinstance(U, RemoteObject)

    logger = getLogger('pymor.algorithms.ei.ei_greedy')
    logger.info('Generating Interpolation Data ...')
    logger.info(f'Using pool of {len(pool)} workers for parallel greedy search')

    interpolation_dofs = np.zeros((0,), dtype=np.int32)
    collateral_basis = pool.apply_only(_parallel_ei_greedy_get_empty, 0, U=U)
    max_errs = []
    triangularity_errs = []

    with pool.push({}) as distributed_data:
        errs, snapshot_counts = zip(
            *pool.apply(_parallel_ei_greedy_initialize, U=U, error_norm=error_norm, copy=copy, data=distributed_data)
        )
        snapshot_count = sum(snapshot_counts)
        cum_snapshot_counts = np.hstack(([0], np.cumsum(snapshot_counts)))
        K = np.eye(snapshot_count)  # matrix s.t. U = U_initial.lincomb(K)
        coefficients = np.zeros((0, snapshot_count))
        max_err_ind = np.argmax(errs)
        initial_max_err = max_err = errs[max_err_ind]

        # main loop
        while True:

            if max_interpolation_dofs is not None and len(interpolation_dofs) >= max_interpolation_dofs:
                logger.info('Maximum number of interpolation DOFs reached. Stopping extension loop.')
                logger.info(f'Final maximum interpolation error with '
                            f'{len(interpolation_dofs)} interpolation DOFs: {max_err}')
                break

            logger.info(f'Maximum interpolation error with {len(interpolation_dofs)} interpolation DOFs: {max_err}')

            if atol is not None and max_err <= atol:
                logger.info('Absolute error tolerance reached! Stopping extension loop.')
                break

            if rtol is not None and max_err / initial_max_err <= rtol:
                logger.info('Relative error tolerance reached! Stopping extension loop.')
                break

            # compute new interpolation dof and collateral basis vector
            new_vec, local_ind = pool.apply_only(_parallel_ei_greedy_get_vector, max_err_ind, data=distributed_data)
            new_dof = new_vec.amax()[0][0]
            if new_dof in interpolation_dofs:
                logger.info(f'DOF {new_dof} selected twice for interpolation! Stopping extension loop.')
                break
            new_dof_value = new_vec.dofs([new_dof])[0, 0]
            if new_dof_value == 0.:
                logger.info(f'DOF {new_dof} selected for interpolation has zero maximum error! '
                            f'Stopping extension loop.')
                break
            new_vec *= 1 / new_dof_value
            interpolation_dofs = np.hstack((interpolation_dofs, new_dof))
            collateral_basis.append(new_vec)
            global_max_err_ind = cum_snapshot_counts[max_err_ind] + local_ind
            coefficients = np.vstack([coefficients, K[global_max_err_ind] / new_dof_value])
            max_errs.append(max_err)

            errs, new_dof_values = zip(
                *pool.apply(_parallel_ei_greedy_update, new_vec=new_vec, new_dof=new_dof, data=distributed_data)
            )
            new_dof_values = np.hstack(new_dof_values)
            K -= (K[global_max_err_ind] / new_dof_value) * new_dof_values[:, np.newaxis]
            max_err_ind = np.argmax(errs)
            max_err = errs[max_err_ind]

    interpolation_matrix = collateral_basis.dofs(interpolation_dofs).T
    triangularity_errors = np.abs(interpolation_matrix - np.tril(interpolation_matrix))
    for d in range(1, len(interpolation_matrix) + 1):
        triangularity_errs.append(np.max(triangularity_errors[:d, :d]))

    if len(triangularity_errs) > 0:
        logger.info(f'Interpolation matrix is not lower triangular with maximum error of {triangularity_errs[-1]}')
        logger.info('')

    if nodal_basis:
        logger.info('Building nodal basis.')
        inv_interpolation_matrix = np.linalg.inv(interpolation_matrix)
        collateral_basis = collateral_basis.lincomb(inv_interpolation_matrix.T)
        coefficients = inv_interpolation_matrix.T @ coefficients
        interpolation_matrix = np.eye(len(collateral_basis))

    data = {'errors': max_errs, 'triangularity_errors': triangularity_errs,
            'coefficients': coefficients, 'interpolation_matrix': interpolation_matrix}

    return interpolation_dofs, collateral_basis, data


def _parallel_ei_greedy_get_empty(U=None):
    return U.empty()


def _parallel_ei_greedy_initialize(U=None, error_norm=None, copy=None, data=None):
    if copy:
        U = U.copy()
    data['U'] = U
    data['error_norm'] = error_norm
    errs = U.norm() if error_norm is None else U.sup_norm() if error_norm == 'sup' else error_norm(U)
    data['max_err_ind'] = max_err_ind = np.argmax(errs)
    return errs[max_err_ind], len(U)


def _parallel_ei_greedy_get_vector(data=None):
    return data['U'][data['max_err_ind']].copy(), data['max_err_ind']


def _parallel_ei_greedy_update(new_vec=None, new_dof=None, data=None):
    U = data['U']
    error_norm = data['error_norm']

    new_dof_values = U.dofs([new_dof])[:, 0]
    U.axpy(-new_dof_values, new_vec)

    errs = U.norm() if error_norm is None else U.sup_norm() if error_norm == 'sup' else error_norm(U)
    data['max_err_ind'] = max_err_ind = np.argmax(errs)
    return errs[max_err_ind], new_dof_values


def _identity(x):
    return x



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""Visualization of grid data using OpenGL.

This module provides a widget for displaying patch plots of
scalar data assigned to 2D-grids using OpenGL. This widget is not
intended to be used directly. Instead, use
:meth:`~pymor.discretizers.builtin.gui.qt.visualize_patch` or
:class:`~pymor.discretizers.builtin.gui.qt.PatchVisualizer`.
"""

import math as m

import numpy as np

from pymor.core.config import config


if config.HAVE_QT and config.HAVE_QTOPENGL and config.HAVE_GL:
    import OpenGL.GL as gl
    from qtpy.QtWidgets import QSizePolicy, QOpenGLWidget
    from qtpy.QtGui import QPainter, QFontMetrics
    from ctypes import c_void_p

    from pymor.core.defaults import defaults
    from pymor.discretizers.builtin.grids.constructions import flatten_grid
    from pymor.discretizers.builtin.grids.referenceelements import triangle, square

    def compile_shader(source, vertex=True):
        """Compile a vertex shader from source."""
        shader_type = gl.GL_VERTEX_SHADER if vertex else gl.GL_FRAGMENT_SHADER
        shader = gl.glCreateShader(shader_type)
        gl.glShaderSource(shader, source)
        gl.glCompileShader(shader)
        # check compilation error
        result = gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS)
        if not result:
            raise RuntimeError(gl.glGetShaderInfoLog(shader))
        return shader

    def link_shader_program(shaders):
        """Create a shader program with from compiled shaders."""
        program = gl.glCreateProgram()
        for shader in shaders:
            gl.glAttachShader(program, shader)
        gl.glLinkProgram(program)
        # check linking error
        result = gl.glGetProgramiv(program, gl.GL_LINK_STATUS)
        if not result:
            raise RuntimeError(gl.glGetProgramInfoLog(program))
        return program

    VS = """
    #version 120
    // Attribute variable that contains coordinates of the vertices.
    attribute vec3 position;
    varying float value;

    void main()
    {
        gl_Position.xy = position.xy;
        gl_Position.z = 0.;
        gl_Position.w = 1.;
        value = position.z;
    }
    """

    FS = """
    #version 120

    uniform sampler1D colormap;

    varying float value;

    void main()
    {
        gl_FragColor = texture1D(colormap, value);
    }
    """

    @defaults('name')
    def colormap_texture(name='viridis'):
        resolution = min(gl.GL_MAX_TEXTURE_SIZE, 1024)
        colormap_id = gl.glGenTextures(1)
        gl.glBindTexture(gl.GL_TEXTURE_1D, colormap_id)
        gl.glTexParameteri(gl.GL_TEXTURE_1D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
        gl.glTexParameteri(gl.GL_TEXTURE_1D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
        gl.glTexParameteri(gl.GL_TEXTURE_1D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE)
        colormap = np.empty((resolution, 4), dtype='f4')
        from matplotlib.pyplot import get_cmap
        try:
            cmap = get_cmap(name)
        except ValueError:
            from pymor.core.logger import getLogger
            # this is our default which might not exist for older matplotlib
            # so a warning would be annoying
            if name != 'viridis':
                msg = f'Unknown colormap {name}, using default colormap'
                getLogger('pymor.discretizers.builtin.gui.gl.colormap_texture').warning(msg)
            cmap = get_cmap()
        colormap[:] = cmap(np.linspace(0., 1., resolution))
        gl.glTexImage1D(gl.GL_TEXTURE_1D, 0, gl.GL_RGBA, resolution, 0, gl.GL_RGBA, gl.GL_FLOAT, colormap)
        gl.glBindTexture(gl.GL_TEXTURE_1D, 0)
        return colormap_id

    class GLPatchWidget(QOpenGLWidget):

        def __init__(self, parent, grid, vmin=None, vmax=None, bounding_box=([0, 0], [1, 1]), codim=2):
            assert grid.reference_element in (triangle, square)
            assert grid.dim == 2
            assert codim in (0, 2)
            super().__init__(parent)
            self.setMinimumSize(300, 300)
            self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))

            subentities, coordinates, entity_map = flatten_grid(grid)

            self.subentities = subentities
            self.entity_map = entity_map
            self.reference_element = grid.reference_element
            self.vmin = vmin
            self.vmax = vmax
            self.bounding_box = bounding_box
            self.codim = codim
            self.update_vbo = False
            bb = self.bounding_box
            self.size = np.array([bb[1][0] - bb[0][0], bb[1][1] - bb[0][1]])
            self.scale = 2 / self.size
            self.shift = - np.array(bb[0]) - self.size / 2

            # setup buffers
            buffer_dtype = [('position', 'f4', 2), ('color', 'f4')]
            if self.reference_element == triangle:
                if codim == 2:
                    self.vertex_data = np.empty(len(coordinates), dtype=buffer_dtype)
                    self.indices = subentities
                else:
                    self.vertex_data = np.empty(len(subentities) * 3, dtype=buffer_dtype)
                    self.indices = np.arange(len(subentities) * 3, dtype=np.uint32)
            else:
                if codim == 2:
                    self.vertex_data = np.empty(len(coordinates), dtype=buffer_dtype)
                    self.indices = np.vstack((subentities[:, 0:3], subentities[:, [0, 2, 3]]))
                else:
                    self.vertex_data = np.empty(len(subentities) * 6, dtype=buffer_dtype)
                    self.indices = np.arange(len(subentities) * 6, dtype=np.uint32)
            self.indices = np.ascontiguousarray(self.indices)

            self.vertex_data['color'] = 1

            self.set_coordinates(coordinates)
            self.set(np.zeros(grid.size(codim)))

        def resizeGL(self, w, h):
            gl.glViewport(0, 0, w, h)
            gl.glLoadIdentity()
            self.update()

        def initializeGL(self):
            gl.glClearColor(1.0, 1.0, 1.0, 1.0)

            self.shaders_program = link_shader_program([compile_shader(VS, vertex=True),
                                                        compile_shader(FS, vertex=False)])
            gl.glUseProgram(self.shaders_program)

            self.vertices_id = gl.glGenBuffers(1)
            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vertices_id)
            gl.glBufferData(gl.GL_ARRAY_BUFFER, self.vertex_data, gl.GL_DYNAMIC_DRAW)
            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)

            self.indices_id = gl.glGenBuffers(1)
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.indices_id)
            gl.glBufferData(gl.GL_ELEMENT_ARRAY_BUFFER, self.indices, gl.GL_STATIC_DRAW)
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)

            self.colormap_id = colormap_texture()
            self.colormap_location = gl.glGetUniformLocation(self.shaders_program, 'colormap')

        def paintGL(self):
            if self.update_vbo:
                gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vertices_id)
                gl.glBufferData(gl.GL_ARRAY_BUFFER, self.vertex_data, gl.GL_DYNAMIC_DRAW)
                gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
                self.update_vbo = False

            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

            gl.glUniform1i(self.colormap_location, 0)
            gl.glActiveTexture(gl.GL_TEXTURE0 + 0)
            gl.glBindTexture(gl.GL_TEXTURE_1D, self.colormap_id)

            gl.glPushClientAttrib(gl.GL_CLIENT_VERTEX_ARRAY_BIT)
            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vertices_id)
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, self.indices_id)

            gl.glVertexPointer(3, gl.GL_FLOAT, 0, c_void_p(None))
            gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
            gl.glDrawElements(gl.GL_TRIANGLES, self.indices.size, gl.GL_UNSIGNED_INT, None)
            gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, 0)
            gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)
            gl.glPopClientAttrib()

        def set_coordinates(self, coordinates):
            if self.codim == 2:
                self.vertex_data['position'][:, 0:2] = coordinates
                self.vertex_data['position'][:, 0:2] += self.shift
                self.vertex_data['position'][:, 0:2] *= self.scale
            elif self.reference_element == triangle:
                VERTEX_POS = coordinates[self.subentities]
                VERTEX_POS += self.shift
                VERTEX_POS *= self.scale
                self.vertex_data['position'][:, 0:2] = VERTEX_POS.reshape((-1, 2))
            else:
                num_entities = len(self.subentities)
                VERTEX_POS = coordinates[self.subentities]
                VERTEX_POS += self.shift
                VERTEX_POS *= self.scale
                self.vertex_data['position'][0:num_entities * 3, 0:2] = VERTEX_POS[:, 0:3, :].reshape((-1, 2))
                self.vertex_data['position'][num_entities * 3:, 0:2] = VERTEX_POS[:, [0, 2, 3], :].reshape((-1, 2))
            self.update_vbo = True
            self.update()

        def set(self, U, vmin=None, vmax=None):
            self.vmin = self.vmin if vmin is None else vmin
            self.vmax = self.vmax if vmax is None else vmax

            U_buffer = self.vertex_data['color']
            if self.codim == 2:
                U_buffer[:] = U[self.entity_map]
            elif self.reference_element == triangle:
                U_buffer[:] = np.repeat(U, 3)
            else:
                U_buffer[:] = np.tile(np.repeat(U, 3), 2)

            # normalize
            vmin = np.min(U) if self.vmin is None else self.vmin
            vmax = np.max(U) if self.vmax is None else self.vmax
            U_buffer -= vmin
            if (vmax - vmin) > 0:
                U_buffer /= float(vmax - vmin)

            self.update_vbo = True
            self.update()

    class ColorBarWidget(QOpenGLWidget):

        def __init__(self, parent, U=None, vmin=None, vmax=None):
            super().__init__(parent)
            self.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding))
            self.set(U, vmin, vmax)

        def resizeGL(self, w, h):
            gl.glViewport(0, 0, w, h)
            gl.glLoadIdentity()
            self.update()

        def initializeGL(self):
            gl.glClearColor(1.0, 1.0, 1.0, 1.0)
            self.shaders_program = link_shader_program([compile_shader(VS, vertex=True),
                                                        compile_shader(FS, vertex=False)])
            gl.glUseProgram(self.shaders_program)

            self.colormap_id = colormap_texture()
            self.colormap_location = gl.glGetUniformLocation(self.shaders_program, 'colormap')

        def set(self, U=None, vmin=None, vmax=None):
            # normalize U
            fm = QFontMetrics(self.font())
            self.vmin = vmin if vmin is not None else (np.min(U) if U is not None else 0.)
            self.vmax = vmax if vmax is not None else (np.max(U) if U is not None else 1.)
            difference = abs(self.vmin - self.vmax)
            if difference == 0:
                precision = 3
            else:
                precision = m.log(max(abs(self.vmin), abs(self.vmax)) / difference, 10) + 1
                precision = int(min(max(precision, 3), 8))
            self.vmin_str = format(('{:.' + str(precision) + '}').format(self.vmin))
            self.vmax_str = format(('{:.' + str(precision) + '}').format(self.vmax))
            self.vmin_width = fm.width(self.vmin_str)
            self.vmax_width = fm.width(self.vmax_str)
            self.text_height = fm.height() * 1.5
            self.text_ascent = fm.ascent() * 1.5
            self.text_descent = fm.descent() * 1.5
            self.setMinimumSize(max(self.vmin_width, self.vmax_width) + 20, 300)
            self.update()

        def paintGL(self):
            p = QPainter(self)
            p.beginNativePainting()
            gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
            gl.glUseProgram(self.shaders_program)
            gl.glUniform1i(self.colormap_location, 0)
            gl.glActiveTexture(gl.GL_TEXTURE0 + 0)
            gl.glBindTexture(gl.GL_TEXTURE_1D, self.colormap_id)

            gl.glBegin(gl.GL_QUAD_STRIP)
            bar_start = -1 + self.text_height / self.height() * 2
            bar_height = (1 - 2 * self.text_height / self.height()) * 2
            steps = 40
            for i in range(steps + 1):
                y = i * (1 / steps)
                # gl.glColor(y, 0, 0)
                gl.glVertex(-0.5, (bar_height*y + bar_start), y)
                gl.glVertex(0.5, (bar_height*y + bar_start), y)
            gl.glEnd()
            p.endNativePainting()
            p.drawText(round((self.width() - self.vmax_width)/2), self.text_ascent, self.vmax_str)
            p.drawText(round((self.width() - self.vmin_width)/2), self.height() - self.text_height + self.text_ascent,
                       self.vmin_str)
            p.end()

else:

    class GLPatchWidget:
        pass

    class ColorBarWidget:
        pass
